@@ -1,40 +1,48 @@
+// Copyright (c) HashiCorp, Inc.
+// SPDX-License-Identifier: BUSL-1.1
+
 package vault
 
 import (
 	"context"
 	"fmt"
+	"regexp"
 	"strings"
 	"sync"
 	"sync/atomic"
 	"time"
 
-	metrics "github.com/armon/go-metrics"
-	radix "github.com/armon/go-radix"
-	"github.com/hashicorp/vault/helper/consts"
+	"github.com/armon/go-metrics"
+	"github.com/armon/go-radix"
+	"github.com/hashicorp/go-hclog"
+	"github.com/hashicorp/go-secure-stdlib/strutil"
 	"github.com/hashicorp/vault/helper/namespace"
-	"github.com/hashicorp/vault/helper/salt"
-	"github.com/hashicorp/vault/helper/strutil"
-	"github.com/hashicorp/vault/logical"
+	"github.com/hashicorp/vault/sdk/helper/consts"
+	"github.com/hashicorp/vault/sdk/helper/salt"
+	"github.com/hashicorp/vault/sdk/logical"
 )
 
-var (
-	denylistHeaders = []string{
-		"Authorization",
-		consts.AuthHeaderName,
-	}
-)
+var deniedPassthroughRequestHeaders = []string{
+	consts.AuthHeaderName,
+}
+
+// matches when '+' is next to a non-slash char
+var wcAdjacentNonSlashRegEx = regexp.MustCompile(`\+[^/]|[^/]\+`).MatchString
 
 // Router is used to do prefix based routing of a request to a logical backend
 type Router struct {
 	l                  sync.RWMutex
 	root               *radix.Tree
+	binaryPaths        *radix.Tree
 	mountUUIDCache     *radix.Tree
 	mountAccessorCache *radix.Tree
 	tokenStoreSaltFunc func(context.Context) (*salt.Salt, error)
 	// storagePrefix maps the prefix used for storage (ala the BarrierView)
 	// to the backend. This is used to map a key back into the backend that owns it.
 	// For example, logical/uuid1/foobar -> secrets/ (kv backend) + foobar
-	storagePrefix *radix.Tree
+	storagePrefix            *radix.Tree
+	logger                   hclog.Logger
+	rollbackMetricsMountName bool
 }
 
 // NewRouter returns a new router
@@ -44,32 +52,104 @@ func NewRouter() *Router {
 		storagePrefix:      radix.New(),
 		mountUUIDCache:     radix.New(),
 		mountAccessorCache: radix.New(),
+		// this will get replaced in production with a real logger but it's useful to have a default in place for tests
+		logger: hclog.NewNullLogger(),
 	}
 	return r
 }
 
 // routeEntry is used to represent a mount point in the router
 type routeEntry struct {
-	tainted       bool
-	backend       logical.Backend
-	mountEntry    *MountEntry
-	storageView   logical.Storage
-	storagePrefix string
-	rootPaths     atomic.Value
-	loginPaths    atomic.Value
-	l             sync.RWMutex
+	tainted atomic.Bool
+	// backend is the actual backend instance for this route entry; lock l must
+	// be held to access this field.
+	backend                logical.Backend
+	mountEntry             *MountEntry
+	storageView            logical.Storage
+	storagePrefix          string
+	rootPaths              atomic.Value
+	loginPaths             atomic.Value
+	binaryPaths            atomic.Value
+	limitedPaths           atomic.Value
+	allowSnapshotReadPaths atomic.Value
+	// l is the lock used to protect access to backend during reloads
+	l sync.RWMutex
 }
 
-type validateMountResponse struct {
+type wildcardPath struct {
+	// this sits in the hot path of requests so we are micro-optimizing by
+	// storing pre-split slices of path segments
+	segments []string
+	isPrefix bool
+}
+
+// specialPathsEntry is used to hold the routeEntry specialPaths
+type specialPathsEntry struct {
+	paths         *radix.Tree
+	wildcardPaths []wildcardPath
+}
+
+// specialPathsLookupFunc is used by (*Router).specialPath to look up a
+// specialPathsEntry corresponding to loginPath, binaryPath, limitedPath, or
+// allowSnapshotReadPath.
+type specialPathsLookupFunc func(re *routeEntry) *specialPathsEntry
+
+type ValidateMountResponse struct {
 	MountType     string `json:"mount_type" structs:"mount_type" mapstructure:"mount_type"`
 	MountAccessor string `json:"mount_accessor" structs:"mount_accessor" mapstructure:"mount_accessor"`
 	MountPath     string `json:"mount_path" structs:"mount_path" mapstructure:"mount_path"`
 	MountLocal    bool   `json:"mount_local" structs:"mount_local" mapstructure:"mount_local"`
 }
 
-// validateMountByAccessor returns the mount type and ID for a given mount
+func (r *Router) reset() {
+	r.l.Lock()
+	defer r.l.Unlock()
+	r.root = radix.New()
+	r.storagePrefix = radix.New()
+	r.mountUUIDCache = radix.New()
+	r.mountAccessorCache = radix.New()
+}
+
+func (r *Router) GetRecords(tag string) ([]map[string]interface{}, error) {
+	r.l.RLock()
+	defer r.l.RUnlock()
+	var data []map[string]interface{}
+	var tree *radix.Tree
+	switch tag {
+	case "root":
+		tree = r.root
+	case "uuid":
+		tree = r.mountUUIDCache
+	case "accessor":
+		tree = r.mountAccessorCache
+	case "storage":
+		tree = r.storagePrefix
+	default:
+		return nil, logical.ErrUnsupportedPath
+	}
+	for _, v := range tree.ToMap() {
+		info := v.(Deserializable).Deserialize()
+		data = append(data, info)
+	}
+	return data, nil
+}
+
+func (entry *routeEntry) Deserialize() map[string]interface{} {
+	entry.l.RLock()
+	defer entry.l.RUnlock()
+	ret := map[string]interface{}{
+		"tainted":        entry.tainted.Load(),
+		"storage_prefix": entry.storagePrefix,
+	}
+	for k, v := range entry.mountEntry.Deserialize() {
+		ret[k] = v
+	}
+	return ret
+}
+
+// ValidateMountByAccessor returns the mount type and ID for a given mount
 // accessor
-func (r *Router) validateMountByAccessor(accessor string) *validateMountResponse {
+func (r *Router) ValidateMountByAccessor(accessor string) *ValidateMountResponse {
 	if accessor == "" {
 		return nil
 	}
@@ -84,7 +164,7 @@ func (r *Router) validateMountByAccessor(accessor string) *validateMountResponse
 		mountPath = credentialRoutePrefix + mountPath
 	}
 
-	return &validateMountResponse{
+	return &ValidateMountResponse{
 		MountAccessor: mountEntry.Accessor,
 		MountType:     mountEntry.Type,
 		MountPath:     mountPath,
@@ -122,14 +202,36 @@ func (r *Router) Mount(backend logical.Backend, prefix string, mountEntry *Mount
 
 	// Create a mount entry
 	re := &routeEntry{
-		tainted:       false,
 		backend:       backend,
 		mountEntry:    mountEntry,
-		storagePrefix: storageView.prefix,
+		storagePrefix: storageView.Prefix(),
 		storageView:   storageView,
 	}
+	re.tainted.Store(mountEntry.Tainted)
 	re.rootPaths.Store(pathsToRadix(paths.Root))
-	re.loginPaths.Store(pathsToRadix(paths.Unauthenticated))
+	loginPathsEntry, err := parseUnauthenticatedPaths(paths.Unauthenticated)
+	if err != nil {
+		return err
+	}
+	re.loginPaths.Store(loginPathsEntry)
+
+	binaryPathsEntry, err := parseUnauthenticatedPaths(paths.Binary)
+	if err != nil {
+		return err
+	}
+	re.binaryPaths.Store(binaryPathsEntry)
+
+	limitedPathsEntry, err := parseUnauthenticatedPaths(paths.Limited)
+	if err != nil {
+		return err
+	}
+	re.limitedPaths.Store(limitedPathsEntry)
+
+	allowSnapshotReadPathsEntry, err := parseUnauthenticatedPaths(paths.AllowSnapshotRead)
+	if err != nil {
+		return err
+	}
+	re.allowSnapshotReadPaths.Store(allowSnapshotReadPathsEntry)
 
 	switch {
 	case prefix == "":
@@ -219,7 +321,7 @@ func (r *Router) Taint(ctx context.Context, path string) error {
 	defer r.l.Unlock()
 	_, raw, ok := r.root.LongestPrefix(path)
 	if ok {
-		raw.(*routeEntry).tainted = true
+		raw.(*routeEntry).tainted.Store(true)
 	}
 	return nil
 }
@@ -236,7 +338,7 @@ func (r *Router) Untaint(ctx context.Context, path string) error {
 	defer r.l.Unlock()
 	_, raw, ok := r.root.LongestPrefix(path)
 	if ok {
-		raw.(*routeEntry).tainted = false
+		raw.(*routeEntry).tainted.Store(false)
 	}
 	return nil
 }
@@ -279,23 +381,30 @@ func (r *Router) MatchingMountByAccessor(mountAccessor string) *MountEntry {
 // MatchingMount returns the mount prefix that would be used for a path
 func (r *Router) MatchingMount(ctx context.Context, path string) string {
 	r.l.RLock()
-	mount := r.matchingMountInternal(ctx, path)
+	mount, _ := r.matchingMountInternal(ctx, path)
 	r.l.RUnlock()
 	return mount
 }
 
-func (r *Router) matchingMountInternal(ctx context.Context, path string) string {
+// MatchingMountAndEntry returns the mount prefix and MountEntry used for a path
+func (r *Router) MatchingMountAndEntry(ctx context.Context, path string) (string, *MountEntry) {
+	r.l.RLock()
+	defer r.l.RUnlock()
+	return r.matchingMountInternal(ctx, path)
+}
+
+func (r *Router) matchingMountInternal(ctx context.Context, path string) (string, *MountEntry) {
 	ns, err := namespace.FromContext(ctx)
 	if err != nil {
-		return ""
+		return "", nil
 	}
 	path = ns.Path + path
 
-	mount, _, ok := r.root.LongestPrefix(path)
+	mount, raw, ok := r.root.LongestPrefix(path)
 	if !ok {
-		return ""
+		return "", nil
 	}
-	return mount
+	return mount, raw.(*routeEntry).mountEntry
 }
 
 // matchingPrefixInternal returns a mount prefix that a path may be a part of
@@ -322,7 +431,7 @@ func (r *Router) matchingPrefixInternal(ctx context.Context, path string) string
 func (r *Router) MountConflict(ctx context.Context, path string) string {
 	r.l.RLock()
 	defer r.l.RUnlock()
-	if exactMatch := r.matchingMountInternal(ctx, path); exactMatch != "" {
+	if exactMatch, _ := r.matchingMountInternal(ctx, path); exactMatch != "" {
 		return exactMatch
 	}
 	if prefixMatch := r.matchingPrefixInternal(ctx, path); prefixMatch != "" {
@@ -336,9 +445,11 @@ func (r *Router) MountConflict(ctx context.Context, path string) string {
 func (r *Router) MatchingStorageByAPIPath(ctx context.Context, path string) logical.Storage {
 	return r.matchingStorage(ctx, path, true)
 }
+
 func (r *Router) MatchingStorageByStoragePath(ctx context.Context, path string) logical.Storage {
 	return r.matchingStorage(ctx, path, false)
 }
+
 func (r *Router) matchingStorage(ctx context.Context, path string, apiPath bool) logical.Storage {
 	ns, err := namespace.FromContext(ctx)
 	if err != nil {
@@ -392,24 +503,21 @@ func (r *Router) MatchingBackend(ctx context.Context, path string) logical.Backe
 	if !ok {
 		return nil
 	}
-	return raw.(*routeEntry).backend
+
+	re := raw.(*routeEntry)
+	re.l.RLock()
+	defer re.l.RUnlock()
+
+	return re.backend
 }
 
 // MatchingSystemView returns the SystemView used for a path
 func (r *Router) MatchingSystemView(ctx context.Context, path string) logical.SystemView {
-	ns, err := namespace.FromContext(ctx)
-	if err != nil {
+	backend := r.MatchingBackend(ctx, path)
+	if backend == nil {
 		return nil
 	}
-	path = ns.Path + path
-
-	r.l.RLock()
-	_, raw, ok := r.root.LongestPrefix(path)
-	r.l.RUnlock()
-	if !ok {
-		return nil
-	}
-	return raw.(*routeEntry).backend.System()
+	return backend.System()
 }
 
 // MatchingStoragePrefixByAPIPath the storage prefix for the given api path
@@ -420,13 +528,13 @@ func (r *Router) MatchingStoragePrefixByAPIPath(ctx context.Context, path string
 	}
 	path = ns.Path + path
 
-	_, prefix, found := r.matchingMountEntryByPath(ctx, path, true)
+	_, prefix, found := r.matchingMountEntryByPath(path, true)
 	return prefix, found
 }
 
 // MatchingAPIPrefixByStoragePath the api path information for the given storage path
 func (r *Router) MatchingAPIPrefixByStoragePath(ctx context.Context, path string) (*namespace.Namespace, string, string, bool) {
-	me, prefix, found := r.matchingMountEntryByPath(ctx, path, false)
+	me, prefix, found := r.matchingMountEntryByPath(path, false)
 	if !found {
 		return nil, "", "", found
 	}
@@ -440,7 +548,7 @@ func (r *Router) MatchingAPIPrefixByStoragePath(ctx context.Context, path string
 	return me.Namespace(), mountPath, prefix, found
 }
 
-func (r *Router) matchingMountEntryByPath(ctx context.Context, path string, apiPath bool) (*MountEntry, string, bool) {
+func (r *Router) matchingMountEntryByPath(path string, apiPath bool) (*MountEntry, string, bool) {
 	var raw interface{}
 	var ok bool
 	r.l.RLock()
@@ -491,30 +599,40 @@ func (r *Router) routeCommon(ctx context.Context, req *logical.Request, existenc
 	}
 	r.l.RUnlock()
 	if !ok {
-		return logical.ErrorResponse(fmt.Sprintf("no handler for route '%s'", req.Path)), false, false, logical.ErrUnsupportedPath
+		return logical.ErrorResponse(fmt.Sprintf("no handler for route %q. route entry not found.", req.Path)), false, false, logical.ErrUnsupportedPath
 	}
 	req.Path = adjustedPath
-	defer metrics.MeasureSince([]string{"route", string(req.Operation),
-		strings.Replace(mount, "/", "-", -1)}, time.Now())
+	if !existenceCheck {
+		metricName := []string{"route", string(req.Operation)}
+		if req.Operation != logical.RollbackOperation || r.rollbackMetricsMountName {
+			metricName = append(metricName, strings.ReplaceAll(mount, "/", "-"))
+		}
+		defer metrics.MeasureSince(metricName, time.Now())
+	}
 	re := raw.(*routeEntry)
 
 	// Grab a read lock on the route entry, this protects against the backend
-	// being reloaded during a request.
-	re.l.RLock()
-	defer re.l.RUnlock()
+	// being reloaded during a request. The exception is a renew request on the
+	// token store; such a request will have already been routed through the
+	// token store -> exp manager -> here so we need to not grab the lock again
+	// or we'll be recursively grabbing it.
+	if !(req.Operation == logical.RenewOperation && strings.HasPrefix(req.Path, "auth/token/")) {
+		re.l.RLock()
+		defer re.l.RUnlock()
+	}
 
 	// Filtered mounts will have a nil backend
 	if re.backend == nil {
-		return logical.ErrorResponse(fmt.Sprintf("no handler for route '%s'", req.Path)), false, false, logical.ErrUnsupportedPath
+		return logical.ErrorResponse(fmt.Sprintf("no handler for route %q. route entry found, but backend is nil.", req.Path)), false, false, logical.ErrUnsupportedPath
 	}
 
 	// If the path is tainted, we reject any operation except for
 	// Rollback and Revoke
-	if re.tainted {
+	if re.tainted.Load() {
 		switch req.Operation {
 		case logical.RevokeOperation, logical.RollbackOperation:
 		default:
-			return logical.ErrorResponse(fmt.Sprintf("no handler for route '%s'", req.Path)), false, false, logical.ErrUnsupportedPath
+			return logical.ErrorResponse(fmt.Sprintf("no handler for route %q. route entry is tainted.", req.Path)), false, false, logical.ErrUnsupportedPath
 		}
 	}
 
@@ -523,12 +641,15 @@ func (r *Router) routeCommon(ctx context.Context, req *logical.Request, existenc
 	req.Path = strings.TrimPrefix(ns.Path+req.Path, mount)
 	req.MountPoint = mount
 	req.MountType = re.mountEntry.Type
+	req.SetMountRunningSha256(re.mountEntry.RunningSha256)
+	req.SetMountRunningVersion(re.mountEntry.RunningVersion)
+	req.SetMountIsExternalPlugin(re.mountEntry.IsExternalPlugin())
+	req.SetMountClass(re.mountEntry.MountClass())
+
 	if req.Path == "/" {
 		req.Path = ""
 	}
 
-	originalEntReq := req.EntReq()
-
 	// Attach the storage view for the request
 	req.Storage = re.storageView
 
@@ -539,8 +660,9 @@ func (r *Router) routeCommon(ctx context.Context, req *logical.Request, existenc
 	clientToken := req.ClientToken
 	switch {
 	case strings.HasPrefix(originalPath, "auth/token/"):
-	case strings.HasPrefix(originalPath, "sys/"):
-	case strings.HasPrefix(originalPath, cubbyholeMountPath):
+	case strings.HasPrefix(originalPath, mountPathSystem):
+	case strings.HasPrefix(originalPath, mountPathIdentity):
+	case strings.HasPrefix(originalPath, mountPathCubbyhole):
 		if req.Operation == logical.RollbackOperation {
 			// Backend doesn't support this and it can't properly look up a
 			// cubbyhole ID so just return here
@@ -558,7 +680,8 @@ func (r *Router) routeCommon(ctx context.Context, req *logical.Request, existenc
 		}
 
 		switch {
-		case te.NamespaceID == namespace.RootNamespaceID && !strings.HasPrefix(req.ClientToken, "s."):
+		case te.NamespaceID == namespace.RootNamespaceID && !strings.HasPrefix(req.ClientToken, consts.LegacyServiceTokenPrefix) &&
+			!strings.HasPrefix(req.ClientToken, consts.ServiceTokenPrefix):
 			// In order for the token store to revoke later, we need to have the same
 			// salted ID, so we double-salt what's going to the cubbyhole backend
 			salt, err := r.tokenStoreSaltFunc(ctx)
@@ -588,18 +711,36 @@ func (r *Router) routeCommon(ctx context.Context, req *logical.Request, existenc
 	originalClientTokenRemainingUses := req.ClientTokenRemainingUses
 	req.ClientTokenRemainingUses = 0
 
-	origMFACreds := req.MFACreds
+	originalMFACreds := req.MFACreds
 	req.MFACreds = nil
 
+	originalControlGroup := req.ControlGroup
+	req.ControlGroup = nil
+
 	// Cache the headers
 	headers := req.Headers
+	req.Headers = nil
+
+	// Cache the saved request SSC token
+	inboundToken := req.InboundSSCToken
+
+	// Ensure that the inbound token we cache in the
+	// request during token creation isn't sent to backends
+	req.InboundSSCToken = ""
 
 	// Filter and add passthrough headers to the backend
 	var passthroughRequestHeaders []string
 	if rawVal, ok := re.mountEntry.synthesizedConfigCache.Load("passthrough_request_headers"); ok {
 		passthroughRequestHeaders = rawVal.([]string)
 	}
-	req.Headers = filteredPassthroughHeaders(headers, passthroughRequestHeaders)
+	var allowedResponseHeaders []string
+	if rawVal, ok := re.mountEntry.synthesizedConfigCache.Load("allowed_response_headers"); ok {
+		allowedResponseHeaders = rawVal.([]string)
+	}
+
+	if len(passthroughRequestHeaders) > 0 {
+		req.Headers = filteredHeaders(headers, passthroughRequestHeaders, deniedPassthroughRequestHeaders)
+	}
 
 	// Cache the wrap info of the request
 	var wrapInfo *logical.RequestWrapInfo
@@ -620,6 +761,11 @@ func (r *Router) routeCommon(ctx context.Context, req *logical.Request, existenc
 		req.Path = originalPath
 		req.MountPoint = mount
 		req.MountType = re.mountEntry.Type
+		req.SetMountRunningSha256(re.mountEntry.RunningSha256)
+		req.SetMountRunningVersion(re.mountEntry.RunningVersion)
+		req.SetMountIsExternalPlugin(re.mountEntry.IsExternalPlugin())
+		req.SetMountClass(re.mountEntry.MountClass())
+
 		req.Connection = originalConn
 		req.ID = originalReqID
 		req.Storage = nil
@@ -638,10 +784,17 @@ func (r *Router) routeCommon(ctx context.Context, req *logical.Request, existenc
 
 		req.EntityID = originalEntityID
 
-		req.MFACreds = origMFACreds
+		req.MFACreds = originalMFACreds
+
+		req.InboundSSCToken = inboundToken
+
+		// Before resetting the tokenEntry, see if an ExternalID was added
+		if req.TokenEntry() != nil && req.TokenEntry().ExternalID != "" {
+			reqTokenEntry.ExternalID = req.TokenEntry().ExternalID
+		}
 
 		req.SetTokenEntry(reqTokenEntry)
-		req.SetEntReq(originalEntReq)
+		req.ControlGroup = originalControlGroup
 	}()
 
 	// Invoke the backend
@@ -650,43 +803,56 @@ func (r *Router) routeCommon(ctx context.Context, req *logical.Request, existenc
 		return nil, ok, exists, err
 	} else {
 		resp, err := re.backend.HandleRequest(ctx, req)
-		if resp != nil &&
-			resp.Auth != nil {
-			// When a token gets renewed, the request hits this path and
-			// reaches token store. Token store delegates the renewal to the
-			// expiration manager. Expiration manager in-turn creates a
-			// different logical request and forwards the request to the auth
-			// backend that had initially authenticated the login request. The
-			// forwarding to auth backend will make this code path hit for the
-			// second time for the same renewal request. The accessors in the
-			// Alias structs should be of the auth backend and not of the token
-			// store. Therefore, avoiding the overwriting of accessors by
-			// having a check for path prefix having "renew". This gets applied
-			// for "renew" and "renew-self" requests.
-			if !strings.HasPrefix(req.Path, "renew") {
-				if resp.Auth.Alias != nil {
-					resp.Auth.Alias.MountAccessor = re.mountEntry.Accessor
-				}
-				for _, alias := range resp.Auth.GroupAliases {
-					alias.MountAccessor = re.mountEntry.Accessor
-				}
+		if resp != nil {
+			if len(allowedResponseHeaders) > 0 {
+				resp.Headers = filteredHeaders(resp.Headers, allowedResponseHeaders, nil)
+			} else {
+				resp.Headers = nil
 			}
 
-			switch re.mountEntry.Type {
-			case "token", "ns_token":
-				// Nothing; we respect what the token store is telling us and
-				// we don't allow tuning
-			default:
-				switch re.mountEntry.Config.TokenType {
-				case logical.TokenTypeService, logical.TokenTypeBatch:
-					resp.Auth.TokenType = re.mountEntry.Config.TokenType
-				case logical.TokenTypeDefault, logical.TokenTypeDefaultService:
-					if resp.Auth.TokenType == logical.TokenTypeDefault {
-						resp.Auth.TokenType = logical.TokenTypeService
+			if resp.Auth != nil {
+				// When a token gets renewed, the request hits this path and
+				// reaches token store. Token store delegates the renewal to the
+				// expiration manager. Expiration manager in-turn creates a
+				// different logical request and forwards the request to the auth
+				// backend that had initially authenticated the login request. The
+				// forwarding to auth backend will make this code path hit for the
+				// second time for the same renewal request. The accessors in the
+				// Alias structs should be of the auth backend and not of the token
+				// store. Therefore, avoiding the overwriting of accessors by
+				// having a check for path prefix having "renew". This gets applied
+				// for "renew" and "renew-self" requests.
+				if !strings.HasPrefix(req.Path, "renew") {
+					if resp.Auth.Alias != nil {
+						resp.Auth.Alias.MountAccessor = re.mountEntry.Accessor
 					}
-				case logical.TokenTypeDefaultBatch:
-					if resp.Auth.TokenType == logical.TokenTypeDefault {
-						resp.Auth.TokenType = logical.TokenTypeBatch
+					for _, alias := range resp.Auth.GroupAliases {
+						alias.MountAccessor = re.mountEntry.Accessor
+					}
+				}
+
+				switch re.mountEntry.Type {
+				case mountTypeToken, mountTypeNSToken:
+					// Nothing; we respect what the token store is telling us and
+					// we don't allow tuning
+				default:
+					switch re.mountEntry.Config.TokenType {
+					case logical.TokenTypeService, logical.TokenTypeBatch:
+						resp.Auth.TokenType = re.mountEntry.Config.TokenType
+					case logical.TokenTypeDefault, logical.TokenTypeDefaultService:
+						switch resp.Auth.TokenType {
+						case logical.TokenTypeDefault, logical.TokenTypeDefaultService, logical.TokenTypeService:
+							resp.Auth.TokenType = logical.TokenTypeService
+						default:
+							resp.Auth.TokenType = logical.TokenTypeBatch
+						}
+					case logical.TokenTypeDefaultBatch:
+						switch resp.Auth.TokenType {
+						case logical.TokenTypeDefault, logical.TokenTypeDefaultBatch, logical.TokenTypeBatch:
+							resp.Auth.TokenType = logical.TokenTypeBatch
+						default:
+							resp.Auth.TokenType = logical.TokenTypeService
+						}
 					}
 				}
 			}
@@ -735,6 +901,45 @@ func (r *Router) RootPath(ctx context.Context, path string) bool {
 
 // LoginPath checks if the given path is used for logins
 func (r *Router) LoginPath(ctx context.Context, path string) bool {
+	return r.specialPath(ctx, path,
+		func(re *routeEntry) *specialPathsEntry {
+			return re.loginPaths.Load().(*specialPathsEntry)
+		})
+}
+
+// BinaryPath checks if the given path is used for binary requests
+func (r *Router) BinaryPath(ctx context.Context, path string) bool {
+	return r.specialPath(ctx, path,
+		func(re *routeEntry) *specialPathsEntry {
+			return re.binaryPaths.Load().(*specialPathsEntry)
+		})
+}
+
+// LimitedPath checks if the given path uses limited requests
+func (r *Router) LimitedPath(ctx context.Context, path string) bool {
+	return r.specialPath(ctx, path,
+		func(re *routeEntry) *specialPathsEntry {
+			return re.limitedPaths.Load().(*specialPathsEntry)
+		})
+}
+
+// AllowSnapshotReadPath checks if the given path is allowed to be used for a
+// snapshot read
+func (r *Router) AllowSnapshotReadPath(ctx context.Context, path string) bool {
+	return r.specialPath(ctx, path,
+		func(re *routeEntry) *specialPathsEntry {
+			return re.allowSnapshotReadPaths.Load().(*specialPathsEntry)
+		})
+}
+
+// specialPath is a common method for checking if the given path has a matching
+// PathsSpecial entry. This is used for Login, Binary, Limited, and
+// AllowSnapshotRead PathsSpecial fields.
+// Matching Priority
+//  1. prefix
+//  2. exact
+//  3. wildcard
+func (r *Router) specialPath(ctx context.Context, path string, lookup specialPathsLookupFunc) bool {
 	ns, err := namespace.FromContext(ctx)
 	if err != nil {
 		return false
@@ -753,25 +958,118 @@ func (r *Router) LoginPath(ctx context.Context, path string) bool {
 	// Trim to get remaining path
 	remain := strings.TrimPrefix(adjustedPath, mount)
 
-	// Check the loginPaths of this backend
-	loginPaths := re.loginPaths.Load().(*radix.Tree)
-	match, raw, ok := loginPaths.LongestPrefix(remain)
-	if !ok {
+	// Check the specialPath of this backend as specified by the caller.
+	pe := lookup(re)
+	match, raw, ok := pe.paths.LongestPrefix(remain)
+	if !ok && len(pe.wildcardPaths) == 0 {
+		// no match found
 		return false
 	}
-	prefixMatch := raw.(bool)
 
-	// Handle the prefix match case
-	if prefixMatch {
-		return strings.HasPrefix(remain, match)
+	if ok {
+		prefixMatch := raw.(bool)
+		if prefixMatch {
+			// Handle the prefix match case
+			return strings.HasPrefix(remain, match)
+		}
+		if match == remain {
+			// Handle the exact match case
+			return true
+		}
 	}
 
-	// Handle the exact match case
-	return match == remain
+	// check Login Paths containing wildcards
+	reqPathParts := strings.Split(remain, "/")
+	for _, w := range pe.wildcardPaths {
+		if pathMatchesWildcardPath(reqPathParts, w.segments, w.isPrefix) {
+			return true
+		}
+	}
+	return false
 }
 
-// pathsToRadix converts a the mapping of special paths to a mapping
-// of special paths to radix trees.
+// pathMatchesWildcardPath returns true if the path made up of the path slice
+// matches the given wildcard path slice
+func pathMatchesWildcardPath(path, wcPath []string, isPrefix bool) bool {
+	if len(wcPath) == 0 {
+		return false
+	}
+
+	if len(path) < len(wcPath) {
+		// check if the path coming in is shorter; if so it can't match
+		return false
+	}
+	if !isPrefix && len(wcPath) != len(path) {
+		// If it's not a prefix we expect the same number of segments
+		return false
+	}
+
+	for i, wcPathPart := range wcPath {
+		switch {
+		case wcPathPart == "+":
+		case wcPathPart == path[i]:
+		case isPrefix && i == len(wcPath)-1 && strings.HasPrefix(path[i], wcPathPart):
+		default:
+			// we encountered segments that did not match
+			return false
+		}
+	}
+	return true
+}
+
+func wildcardError(path, msg string) error {
+	return fmt.Errorf("path %q: invalid use of wildcards %s", path, msg)
+}
+
+func isValidUnauthenticatedPath(path string) (bool, error) {
+	switch {
+	case strings.Count(path, "*") > 1:
+		return false, wildcardError(path, "(multiple '*' is forbidden)")
+	case strings.Contains(path, "+*"):
+		return false, wildcardError(path, "('+*' is forbidden)")
+	case strings.Contains(path, "*") && path[len(path)-1] != '*':
+		return false, wildcardError(path, "('*' is only allowed at the end of a path)")
+	case wcAdjacentNonSlashRegEx(path):
+		return false, wildcardError(path, "('+' is not allowed next to a non-slash)")
+	}
+	return true, nil
+}
+
+// parseUnauthenticatedPaths converts a list of special paths to a
+// specialPathsEntry
+func parseUnauthenticatedPaths(paths []string) (*specialPathsEntry, error) {
+	var tempPaths []string
+	tempWildcardPaths := make([]wildcardPath, 0)
+	for _, path := range paths {
+		if ok, err := isValidUnauthenticatedPath(path); !ok {
+			return nil, err
+		}
+
+		if strings.Contains(path, "+") {
+			// Paths with wildcards are not stored in the radix tree because
+			// the radix tree does not handle wildcards in the middle of strings.
+			isPrefix := false
+			if path[len(path)-1] == '*' {
+				isPrefix = true
+				path = path[0 : len(path)-1]
+			}
+			// We are micro-optimizing by storing pre-split slices of path segments
+			wcPath := wildcardPath{segments: strings.Split(path, "/"), isPrefix: isPrefix}
+			tempWildcardPaths = append(tempWildcardPaths, wcPath)
+		} else {
+			// accumulate paths that do not contain wildcards
+			// to be stored in the radix tree
+			tempPaths = append(tempPaths, path)
+		}
+	}
+
+	return &specialPathsEntry{
+		paths:         pathsToRadix(tempPaths),
+		wildcardPaths: tempWildcardPaths,
+	}, nil
+}
+
+// pathsToRadix converts a list of special paths to a radix tree.
 func pathsToRadix(paths []string) *radix.Tree {
 	tree := radix.New()
 	for _, path := range paths {
@@ -787,34 +1085,35 @@ func pathsToRadix(paths []string) *radix.Tree {
 	return tree
 }
 
-// filteredPassthroughHeaders returns a headers map[string][]string that
-// contains the filtered values contained in passthroughHeaders. Filtering of
-// passthroughHeaders from the origHeaders is done is a case-insensitive manner.
-// Headers that match values from denylistHeaders will be ignored.
-func filteredPassthroughHeaders(origHeaders map[string][]string, passthroughHeaders []string) map[string][]string {
-	retHeaders := make(map[string][]string)
-
+// filteredHeaders returns a headers map[string][]string that
+// contains the filtered values contained in candidateHeaders. Filtering of
+// candidateHeaders from the origHeaders is done is a case-insensitive manner.
+// Headers that match values from deniedHeaders will be ignored.
+func filteredHeaders(origHeaders map[string][]string, candidateHeaders, deniedHeaders []string) map[string][]string {
 	// Short-circuit if there's nothing to filter
-	if len(passthroughHeaders) == 0 {
-		return retHeaders
+	if len(candidateHeaders) == 0 {
+		return nil
 	}
 
-	// Filter passthroughHeaders values through denyListHeaders first. Returns the
-	// lowercased the complement set.
-	passthroughHeadersSubset := strutil.Difference(passthroughHeaders, denylistHeaders, true)
+	retHeaders := make(map[string][]string, len(origHeaders))
+
+	// Filter candidateHeaders values through deniedHeaders first. Returns the
+	// lowercased complement set. We call even if no denied headers to get the
+	// values lowercased.
+	allowedCandidateHeaders := strutil.Difference(candidateHeaders, deniedHeaders, true)
 
 	// Create a map that uses lowercased header values as the key and the original
 	// header naming as the value for comparison down below.
-	lowerHeadersRef := make(map[string]string, len(origHeaders))
+	lowerOrigHeaderKeys := make(map[string]string, len(origHeaders))
 	for key := range origHeaders {
-		lowerHeadersRef[strings.ToLower(key)] = key
+		lowerOrigHeaderKeys[strings.ToLower(key)] = key
 	}
 
 	// Case-insensitive compare of passthrough headers against originating
 	// headers. The returned headers will be the same casing as the originating
 	// header name.
-	for _, ph := range passthroughHeadersSubset {
-		if header, ok := lowerHeadersRef[ph]; ok {
+	for _, ch := range allowedCandidateHeaders {
+		if header, ok := lowerOrigHeaderKeys[ch]; ok {
 			retHeaders[header] = origHeaders[header]
 		}
 	}
